class Theme:
    panel_title = "bold yellow"
    heading_h1 = "bold bright_green on dark_green"
    heading_h2 = "bold yellow"
    table_field = "bold bright_magenta"
    table_value = "not bold default"
    popularity_source = "bright_cyan"
    inline_stat = "cyan"
    vendor_list = "cyan"
    nameserver_list = "bright_blue"
    disclaimer = "italic white on red"
    tags = "bright_white on black"
    tags_green = "bright_green on black"
    tags_red = "bright_red on black"
    product = "orange_red1"
    port = "bright_cyan"
    transport = "cyan"
    footer = "cyan"
    info = "bold green"
    warn = "bold yellow"
    error = "bold red"
    timestamp_date = "not bold default"
    timestamp_t = "dim bright_white"
    timestamp_time = "dim default"
    timestamp_z = "dim bright_white"
    asn_org = "bright_white"
    whois_org = "bright_cyan"
    urlhaus_bl_name = "bright_cyan"
    urlhaus_bl_low = "bright_white on green"
    urlhaus_bl_med = "black on yellow"
    urlhaus_bl_high = "bright_white on red"
